# coding:utf-8

from tweb.error_exception import ErrException, ERROR
from tweb import time
from webmother.utils.bson_util import bson_id, json_id
from webmother.service import ctrl_catalog
from webmother.passport import Passport
from .internal import ethereum as in_ethereum
from .external import ethereum as ex_ethereum
from ..db import mongo as db


class Status:
    removed = -10  # 已删除
    sleeping = 0  # 休眠中
    activated = 10  # 活动中

    default = activated

    status_map = {
        sleeping: {'activate': activated, 'remove': removed},
        activated: {'deactivate': sleeping}
    }

    @staticmethod
    def trans(cur_status, action):
        """
        在当前状态，进行操作将会得到新的状态
        :param cur_status: 当前状态
        :param action: 操作名称
        :return: 新的状态
        """

        valid_actions = Status.status_map.get(cur_status)
        if valid_actions is None:
            raise ErrException(ERROR.E40022, extra=f'current status is {cur_status}, forbid change status')

        new_status = valid_actions.get(action)
        if new_status is None:
            raise ErrException(ERROR.E40022, extra=f'current status is {cur_status}, wrong action [{action}]')

        return new_status


def query_contracts(cid, page_no, page_size, *auth_args):
    """
    查询某区块链下已注册的合约列表
    """
    # c = ctrl_catalog.simple_read(cid)

    # 授权检查
    # Passport().verify(*auth_args).operable('{}/*'.format(c.get('node')), 'contract.read')
    # END

    skip = (page_no - 1) * page_size
    cursor = db.contract.find({
        'catalog': bson_id(cid),
        'status': {'$gte': 0}
    }, {'catalog': 0}).skip(skip).limit(page_size)

    array = list()
    for item in cursor:
        item[f'contract_id'] = json_id(item.pop('_id'))
        array.append(item)
    return array, cursor.count()


def register(cid, data, *auth_args):
    c = ctrl_catalog.simple_read(cid)

    # 授权检查
    Passport().verify(*auth_args).operable('{}/*'.format(c.get('node')), 'contract.create')
    # END

    chain_name = c['name']
    if chain_name not in ['eth']:
        raise ErrException(ERROR.E40000, 'unknown chain name % about id %s' % (chain_name, cid))

    supported = ['main', 'ERC20']

    t = data['type']
    if t not in supported:
        raise ErrException(ERROR.E40000, extra='only support type of %s' % supported)

    now = time.millisecond()
    w = {
        "catalog": bson_id(cid),
        "icon": data.get('icon'),
        "type": t,
        "status": Status.default,
        "created": now,
        "updated": now
    }
    if 'icon' in data:
        w['icon'] = data['icon']
    if t != 'main':
        # 是合约，则增加合约spec
        spec = data['spec']
        if spec is None:
            raise ErrException(ERROR.E40000, extra='no spec field')
        contract_address = spec.get('address').lower()
        if not in_ethereum.is_address(contract_address):
            raise ErrException(ERROR.E40000, extra='invalid contract address')

        if chain_name == 'eth':
            abi = spec.get('data')
            if abi is None:
                raise ErrException(ERROR.E40000, extra='need abi for eth')
            # 校验是否为合法合约参数
            if not in_ethereum.get_contract(contract_address, abi):
                raise ErrException(ERROR.E40000, extra='wrong contract params')

        w['spec'] = {
            'address': contract_address,
            'data': spec.get('data')
        }

        # 保证合约地址不重复
        existed = db.contract.find_one({'spec.address': contract_address, 'status': {'$gte': 0}})
        if existed is not None:
            raise ErrException(ERROR.E40020, extra='existed about the contract address: %s' % contract_address)

        w['name'] = ex_ethereum.contract_call(w, 'name', [])
        w['symbol'] = ex_ethereum.contract_call(w, 'symbol', []).upper()
        w['decimals'] = int(ex_ethereum.contract_call(w, 'decimals', []))

    else:
        if chain_name == 'eth':
            w['name'] = 'Ethereum'
            w['symbol'] = 'ETH'
            w['decimals'] = 18

        existed = db.contract.find_one({'catalog': bson_id(cid), 'type': t, 'status': {'$gte': 0}},
                                       {'catalog': 0})
        if existed is not None:
            existed[f'contract_id'] = json_id(existed.pop('_id'))
            return existed

    result = db.contract.insert_one(w)
    ret = db.contract.find_one(result.inserted_id, {'catalog': 0})
    ret[f'contract_id'] = json_id(ret.pop('_id'))

    return ret


def unregister(cid, contract_id, *auth_args):
    c = ctrl_catalog.simple_read(cid)

    # 授权检查
    Passport().verify(*auth_args).operable('{}/*'.format(c.get('node')), 'contract.remove')
    # END

    db.contract.update_one({'_id': bson_id(contract_id)}, {'$set': {'status': Status.removed}})

    return {}


def read(cid, contract_id, *auth_args):
    # c = ctrl_catalog.simple_read(cid)

    # 授权检查
    # Passport().verify(*auth_args).operable('{}/*'.format(c.get('node')), 'contract.read')
    # END

    return simple_read(contract_id)


def simple_read(contract_id):
    ret = db.contract.find_one({'_id': bson_id(contract_id)})
    if ret is None:
        raise ErrException(ERROR.E40400, extra='eth contract not registered: %s' % contract_id)

    ret[f'contract_id'] = json_id(ret.pop('_id'))
    ret['catalog'] = ctrl_catalog.simple_read(ret['catalog'])

    return ret


def get_main_of_chain(cid):
    existed = db.contract.find_one({'catalog': bson_id(cid), 'type': 'main', 'status': {'$gte': 0}},
                                   {'catalog': 0})
    if existed is None:
        raise ErrException(ERROR.E40400, extra='no main contract registered of %s' % cid)

    existed[f'contract_id'] = json_id(existed.pop('_id'))
    return existed


def get_main_of_contract(contract_id):
    contract = simple_read(contract_id)
    cid = contract['catalog']['cid']
    existed = db.contract.find_one({'catalog': bson_id(cid), 'type': 'main', 'status': {'$gte': 0}},
                                   {'catalog': 0})
    if existed is None:
        raise ErrException(ERROR.E40400, extra='no main contract registered of %s' % cid)

    existed[f'contract_id'] = json_id(existed.pop('_id'))
    return existed



# coding:utf-8

from hexbytes import HexBytes
from eth_abi import decode_abi
from tornado import httpclient
import logging
import json
import re
from tweb.error_exception import ErrException, ERROR
from tweb import time
from tweb import tools
from webmother.utils.bson_util import bson_id, json_id
from webmother.service import ctrl_catalog
from . import ctrl_contract
from .internal import ethereum as in_ethereum
from .external import ethereum as ex_ethereum
from ..db import mongo as db
from ..utils import secret
from ..utils.coinnum import CoinNum


async def query_mine(*auth_args):
    uid = auth_args[0]

    cursor = db.wallet.find({'uid': bson_id(uid), 'status': {'$gte': 0}}, {'uid': 0, 'catalog': 0, 'secret': 0})

    array = list()
    for item in cursor:
        item['wallet_id'] = json_id(item.pop('_id'))
        item['contract'] = ctrl_contract.simple_read(json_id(item.pop('contract_id')))
        array.append(item)

    return array


async def create(contract_id, data, *auth_args):
    password = secret.verify_msg(data)
    uid = auth_args[0]

    now = time.millisecond()
    w = {
        "uid": bson_id(uid),
        "contract_id": bson_id(contract_id),
        "status": 10,
        "created": now,
        "updated": now
    }

    if 'alias' in data:
        w['alias'] = data['alias']
    if 'mark' in data:
        w['mark'] = data['mark']

    contract = ctrl_contract.simple_read(contract_id)
    cid = contract['catalog']['cid']
    w['catalog'] = bson_id(cid)

    t = contract['type']
    w['type'] = t

    if t == 'main':
        # 保证只有一个主币账户
        existed = db.wallet.find_one({'uid': bson_id(uid), 'catalog': bson_id(cid), 'status': {'$gte': 0}},
                                     {'uid': 0, 'catalog': 0})
        if existed is not None:
            existed['wallet_id'] = json_id(existed.pop('_id'))
            existed['contract'] = ctrl_contract.simple_read(json_id(existed.pop('contract_id')))
            return existed

        if not re.match(r'^.{6,20}$', password):
            raise ErrException(ERROR.E40000, extra='password should be 6-20 characters')

        chain_name = contract['catalog']['name']
        if chain_name == 'eth':
            w['address'] = in_ethereum.new_account(password)
        else:
            raise ErrException(ERROR.E40000, extra='invalid block-chain: %s' % chain_name)

        w['secret'] = tools.gen_sha256(password + w['address'])
    else:
        existed = db.wallet.find_one(
            {'uid': bson_id(uid), 'contract_id': bson_id(contract_id), 'status': {'$gte': 0}},
            {'uid': 0, 'catalog': 0})
        if existed is not None:
            existed['wallet_id'] = json_id(existed.pop('_id'))
            existed['contract'] = ctrl_contract.simple_read(json_id(existed.pop('contract_id')))
            return existed

        # 保证已经存在主币账户
        tmp = db.wallet.find_one({'uid': bson_id(uid), 'catalog': bson_id(cid), 'status': {'$gte': 0}},
                                 {'address': 1})
        if tmp is None:
            raise ErrException(ERROR.E40000, extra='please create main account first')

        w['address'] = tmp['address']

        # 校验主币钱包密码
        await verify_password(w['address'], password)

    result = db.wallet.insert_one(w)
    ret = db.wallet.find_one(result.inserted_id, {'uid': 0, 'catalog': 0, 'secret': 0})
    ret['wallet_id'] = json_id(ret.pop('_id'))
    ret['contract'] = ctrl_contract.simple_read(json_id(ret.pop('contract_id')))

    return ret


async def remove(contract_id, wallet_id, *auth_args):
    uid = auth_args[0]

    contract = ctrl_contract.simple_read(contract_id)

    t = contract['type']
    if t == 'main':
        # 主币钱包涉及私钥等重要信息，故只做逻辑删除。同时将关联的合约账户全部删除

        existed = db.wallet.find_one({'_id': bson_id(wallet_id),
                                      'uid': bson_id(uid),
                                      'contract_id': bson_id(contract_id),
                                      'status': {'$gte': 0}
                                      })
        if existed is None:
            raise ErrException(ERROR.E40400)

        address = existed['address']

        db.wallet.update_one({'_id': bson_id(wallet_id),
                              'uid': bson_id(uid),
                              'contract_id': bson_id(contract_id)
                              },
                             {'$set': {'status': -10}})

        cursor = db.wallet.find({'address': address.lower(), 'uid': bson_id(uid), 'status': {'$gte': 0}})
        for item in cursor:
            db.wallet.delete_one({'_id': item['_id']})
    else:
        # 合约账户实际是一种方便查看管理的绑定关系，故直接删除

        db.wallet.delete_one({'_id': bson_id(wallet_id), 'uid': bson_id(uid), 'contract_id': bson_id(contract_id)})

    return {}


async def balance(contract_id, address):
    contract = ctrl_contract.simple_read(contract_id)

    decimals = contract['decimals']
    precision = contract.get('precision', decimals)

    chain_name = contract['catalog']['name']
    t = contract['type']
    if chain_name == 'eth':
        if t == 'main':
            amount = ex_ethereum.wallet_balance(address)
        elif t == 'ERC20':
            amount = ex_ethereum.contract_call(contract, 'balanceOf', [address])
        else:
            raise ErrException(ERROR.E40000, extra='invalid contract type: %s' % t)
    else:
        raise ErrException(ERROR.E40000, extra='invalid block-chain: %s' % chain_name)

    return {
        'balance': CoinNum(amount).set(decimals, precision).group()
    }


async def view(contract_id, data):
    """
    合约查看（不用签名，不用矿工费，不会写入数据）
    """
    contract = ctrl_contract.simple_read(contract_id)

    method = data['method']
    if 'params' in data:
        params = data['params']
    else:
        params = list()

    chain_name = contract['catalog']['name']
    if chain_name == 'eth':
        result = ex_ethereum.contract_call(contract, method, params)
        return {
            'result': result
        }
    else:
        raise ErrException(ERROR.E40000, extra='invalid block-chain: %s' % chain_name)


def tx_estimate(contract_id, sender_address, data):
    # secret.verify_msg(data)

    tx, contract = tx_create(contract_id, sender_address, data)

    main_contract = ctrl_contract.get_main_of_chain(contract['catalog']['cid'])

    decimals = main_contract['decimals']
    precision = main_contract.get('precision', decimals)

    amount = tx['gas'] * tx['gasPrice']
    return {
        'amount': CoinNum(amount).set(decimals, precision),
        'symbol': main_contract['symbol'].upper()
    }


async def transfer_miner_fei(contract_id, sender_address, data):
    # secret.verify_msg(data)

    data['method'] = 'transfer'
    data['params'] = [data['to'], int(data['value'])]

    mf = tx_estimate(contract_id, sender_address, {
        'method': 'transfer',
        'params': [data['to'], int(data['value'])]
    })
    mf['amount'] = mf['amount'].group()
    return mf


async def transfer_apply(contract_id, sender_address, data):
    """
    转账交易广播（属交易的一种，是tx_cast方法的特例）
    """
    secret_pwd = data.get('password')
    password = secret.verify_msg(data)
    # 校验密码
    await verify_password(sender_address, password)

    if 'to' not in data:
        raise ErrException(ERROR.E40000, extra='not to field')

    if 'value' not in data:
        raise ErrException(ERROR.E40000, extra='not value field')

    data['method'] = 'transfer'
    data['params'] = [data.pop('to'), data.pop('value')]

    data['password'] = secret_pwd
    if 'message_sign' in data:
        data.pop('message_sign')
    if 'message_ts' in data:
        data.pop('message_ts')

    return await tx_apply(contract_id, sender_address, data)


async def tx_apply(contract_id, sender_address, data):
    """
    通用交易广播
    """
    secret_pwd = data.get('password')
    password = secret.verify_msg(data)
    # 校验密码
    await verify_password(sender_address, password)

    # 参数校验
    contract = ctrl_contract.simple_read(contract_id)

    contract_type = contract['type']
    chain_name = contract['catalog']['name']
    if chain_name not in ['eth']:
        raise ErrException(ERROR.E40000, extra='invalid block-chain: %s' % chain_name)

    if 'params' not in data:
        raise ErrException(ERROR.E40000, extra='not params field')

    if contract_type == 'main':
        pass
    elif contract_type == 'ERC20':
        if 'method' not in data:
            raise ErrException(ERROR.E40000, extra='not method field')
    else:
        raise ErrException(ERROR.E40000, extra='invalid contract type: %s' % contract_type)
    # END

    raw_data = data.copy()
    raw_data['password'] = secret_pwd

    tx_id = bson_id()
    db.tx.insert_one({
        '_id': tx_id,
        'status': 0,
        'contract_id': contract_id,
        'from': sender_address,
        'raw_data': raw_data,
        'ts_apply': time.millisecond()
    })
    return read_tx(tx_id)


def read_tx(tx_id):
    tx = db.tx.find_one({'_id': bson_id(tx_id)}, {'contract_id': 0, 'raw_data': 0})
    tx['tx_id'] = json_id(tx.pop('_id'))
    return tx


async def tx_cast(contract_id, sender_address, data, nonce=None):
    """
    通用交易广播
    """
    password = secret.verify_msg(data)

    raw_tx, contract = tx_create(contract_id, sender_address, data, nonce=nonce)
    # 广播交易
    signed = in_ethereum.sign_tx(sender_address, raw_tx, password)
    tx_hash = ex_ethereum.send_raw_tx(signed['rawTransaction'])
    # 返回交易详情
    cid = contract['catalog']['cid']
    info = await tx_info(cid, tx_hash)
    return info


def tx_create(contract_id, sender_address, data, nonce=None):
    contract = ctrl_contract.simple_read(contract_id)

    contract_type = contract['type']
    chain_name = contract['catalog']['name']
    if chain_name not in ['eth']:
        raise ErrException(ERROR.E40000, extra='invalid block-chain: %s' % chain_name)

    params = data.get('params')
    if 'params' is None:
        raise ErrException(ERROR.E40000, extra='not params field')

    # 如果参数中有地址，则进行转换
    params = params.copy()
    for i, v in enumerate(params):
        if in_ethereum.is_address(v):
            params[i] = in_ethereum.to_crc_address(v)
        if isinstance(params[i], str) and str.isdecimal(params[i]):
            params[i] = int(params[i])

    if contract_type == 'main':
        tx = ex_ethereum.create_tx(sender_address, params[0], params[1])
    elif contract_type == 'ERC20':
        method = data.get('method')
        if 'method' is None:
            raise ErrException(ERROR.E40000, extra='not method field')

        if nonce is None:
            nonce = ex_ethereum.get_tx_count(sender_address)
        gas_price = ex_ethereum.get_gas_price()
        tx = ex_ethereum.contract_create_tx(contract, method, params, {
            'from': in_ethereum.to_crc_address(sender_address),
            'gasPrice': gas_price,
            'nonce': nonce
        })
    else:
        raise ErrException(ERROR.E40000, extra='invalid contract type: %s' % contract_type)

    return tx, contract


async def tx_list(cid, address):
    address = address.lower()

    c = ctrl_catalog.simple_read(cid)
    chain_type = c['name']

    cond = {'chain_type': chain_type,
            'chain_name': ex_ethereum.chain,
            '$or': [
                {'from': address},
                {'to': address}
            ]}
    if chain_type == 'eth':
        the_last = db.transaction.find_one(cond, {'_id': 0, 'blockNumber': 1, 'synced_block_number': 1, 'synced': 1})
        if the_last is None:
            synced_block_number = 0
            synced = 0
        else:
            synced = the_last.get('synced', 0)
            if 'synced_block_number' not in the_last:
                synced_block_number = the_last['blockNumber']
            else:
                synced_block_number = the_last['synced_block_number']
        block_number = ex_ethereum.block_number()

        now = time.millisecond()
        if ex_ethereum.chain == 'main' and block_number > synced_block_number and now - synced > 30000:
            start = synced_block_number + 1
            http_client = httpclient.AsyncHTTPClient()
            try:
                url = 'http://api.etherscan.io/api?module=account&action=txlist&address={}&startblock={}&' \
                      'sort=desc&apikey=AHCCB5IKCR1RM49ZC5R2TD453DUHVC747V'
                url = url.format(address, start)
                response = await http_client.fetch(url)
                response = json.loads(response.body)

                if response['status'] == '1':
                    items = response['result']
                    for item in items:
                        item['chain_type'] = chain_type
                        item['chain_name'] = ex_ethereum.chain
                        if 'from' in item:
                            item['from'] = item['from'].lower()
                        if 'to' in item:
                            item['to'] = item['to'].lower()
                        if 'blockNumber' in item:
                            item['blockNumber'] = int(item['blockNumber'])
                        if 'transactionIndex' in item:
                            item['transactionIndex'] = int(item['transactionIndex'])
                        if 'timeStamp' in item:
                            item['timeStamp'] = int(item['timeStamp'])
                        if 'confirmations' in item:
                            item.pop('confirmations')

                        try:
                            db.transaction.insert_one(item)
                        except Exception:
                            pass
                elif response['status'] == '0' and response['message'] == 'No transactions found':
                    if the_last is not None:
                        temp = {
                            # 减10目的是为了防止etherscan服务同步链上数据时间误差
                            'synced_block_number': max(block_number - 10, synced_block_number),
                            'synced': now
                        }
                        db.transaction.update_one(cond, {'$set': temp})
                else:
                    raise ErrException(ERROR.E50000, extra='error from etherscan: %s' % response['message'])

            except httpclient.HTTPError as e:
                logging.exception(e)
                raise ErrException(ERROR.E50000, extra=e)
            except Exception as e:
                raise ErrException(ERROR.E50000, extra=e)
            finally:
                http_client.close()

        cursor = db.transaction.find(cond, {
            '_id': 0, 'chain_type': 0, 'chain_name': 0, 'synced_block_number': 0, 'synced': 0
        })
        return list(cursor)
    else:
        raise ErrException(ERROR.E40000, extra='invalid block-chain: %s' % chain_type)


async def tx_details(tx_id):
    tx = db.tx.find_one({'_id': bson_id(tx_id)})
    tx['id'] = json_id(tx.pop('_id'))
    return tx


async def tx_info(cid, tx_hash):
    c = ctrl_catalog.simple_read(cid)
    chain_name = c['name']

    if chain_name == 'eth':
        tx = ex_ethereum.get_tx(tx_hash)

        sender = tx.pop('from')
        ret = {
            'hash': tx.pop('hash'),
            'blockHash': tx.pop('blockHash'),
            'blockNumber': tx.pop('blockNumber'),
            'gas': str(tx.pop('gas')),
            'gasPrice': str(tx.pop('gasPrice')),
            'nonce': tx.pop('nonce'),
            'sender': sender,
            'from': sender
        }

        if tx['input'] == '0x':
            ret['to'] = tx.pop('to')
            ret['value'] = str(tx.pop('value'))

            cnt = db.contract.find_one({'catalog': bson_id(cid), 'type': 'main'})
            ret['symbol'] = cnt['symbol']
            ret['type'] = cnt['type']
            ret['contract_id'] = json_id(cnt['_id'])
        else:
            # 此交易是合约交易
            contract_address = tx.pop('to')

            cnt = db.contract.find_one({'spec.address': contract_address.lower()})
            if cnt is None:
                raise ErrException(ERROR.E50000, extra='unknown contract address: %s' % contract_address)
            contract = in_ethereum.get_contract(contract_address, cnt['spec']['data'])
            abi_result = decode_abi_result(contract, tx['input'])

            contract_type = cnt['type']
            if contract_type == 'ERC20':
                method = abi_result['method']
                params = abi_result['params']
                # 将参数中数字转换为str，太大的整数记入DB会报错
                for i, v in enumerate(params):
                    if isinstance(params[i], int):
                        params[i] = str(params[i])

                if method == 'transfer' and len(params) >= 2:
                    ret['to'] = params[0]
                    ret['value'] = params[1]
                elif method == 'transferFrom' and len(params) >= 3:
                    ret['from'] = params[0]
                    ret['to'] = params[1]
                    ret['value'] = params[2]
                elif method == 'loan' and len(params) >= 3:
                    ret['to'] = params[0]
                    ret['value'] = params[1]
                    ret['charge'] = params[2]
                elif method == 'repay' and len(params) >= 2:
                    ret['to'] = contract_address
                    ret['value'] = params[0]
                    ret['charge'] = params[1]

            ret['symbol'] = cnt['symbol']
            ret['type'] = contract_type
            ret['contract_id'] = json_id(cnt['_id'])
            ret['contract_address'] = contract_address

            tx['abi_result'] = abi_result

        ret['extra'] = tx
        return ret
    else:
        raise ErrException(ERROR.E40000, extra='invalid block-chain: %s' % chain_name)


def decode_abi_result(contract, data):
    data = HexBytes(data)
    selector, params = data[:4], data[4:]
    func = contract.get_function_by_selector(selector)
    types = [x['type'] for x in func.abi['inputs']]
    decoded = decode_abi(types, params)
    return {
        'method': func.fn_name,
        'params': list(decoded)
    }


async def verify_password(address, password):
    existed = db.wallet.find_one({'address': address.lower(), 'type': 'main', 'status': {'$gte': 0}})
    if existed is None:
        raise ErrException(ERROR.E40400, extra='not found the wallet about address')
    if 'secret' in existed:
        secret_hash = tools.gen_sha256(password + address.lower())
        if secret_hash != existed['secret']:
            raise ErrException(ERROR.E40000, extra='wrong password')
    else:
        await in_ethereum.check_pwd(address, password)

        secret_hash = tools.gen_sha256(password + address.lower())
        db.wallet.update_one({'_id': existed['_id']},
                             {'$set': {'secret': secret_hash}})




# coding:utf-8

from config import MongoServer
import pymongo
from pymongo import ASCENDING, DESCENDING

MongoCfg = {
    'authSource': 'wtdb',
    'username': 'app',
    'password': 'Wllt2app',
    'connecttimeoutms': 60 * 1000
}

# wallet DB
mongo_client = None
mongo_db = None

wallet = None
contract = None
# 交易记录
transaction = None
tx = None


def init():
    if not MongoServer['active']:
        return

    global mongo_client
    global mongo_db

    global wallet
    global contract
    global transaction
    global tx

    if mongo_client is not None:
        return

    server = MongoServer['mongodb'].split(':')
    host = server[0]
    port = int(server[1]) if len(server) > 1 else 27017

    mongo_client = pymongo.MongoClient(host=host, port=port, **MongoCfg)
    mongo_db = mongo_client[MongoCfg['authSource']]

    # my collections
    wallet = mongo_db.wallet
    contract = mongo_db.contract
    transaction = mongo_db.transaction
    tx = mongo_db.tx
    # END

    # 创建索引
    _wallet_index()
    _contract_index()
    _transaction_index()
    _tx_index()

    # 初始化系统数据
    _init_data()


def start_session():
    return mongo_client.start_session()


def _wallet_index():
    """
    {
        "_id": ObjectId('5c729df2e155ac16da86a1d0'),
        "uid": ObjectId('5c9c25cae155acde223d5472'),                # 用户ID
        "catalog": ObjectId('5c710622e155ac0c39c8b66d'),            # 所属分类节点ID, 即所属区块链
        "address": '0x58a81beab5f9948114d267ee834d82928accb030',    # 钱包地址
        "contract_id": ObjectId('5c729df2e155ac16da86a1d0'),        # 合约ID，也是币种定义
        "type": "ERC20",     # 币种类型，如：main：主币（本系统将主币也作为一种特殊的合约），ERC20：ERC20币，ERC721：ERC721币
        "alias": '理财钱包',
        "mark": '我不想当韭菜',
        "status": 10,          # -10:已删除，0: 休眠中，10: 活动中
        "created": 1551015331186,
        "updated": 1551015331186
    }
    """
    wallet.create_index('address')
    wallet.create_index([('uid', ASCENDING), ('created', ASCENDING)])
    wallet.create_index([('contract_id', ASCENDING), ('created', DESCENDING)])


def _contract_index():
    """
    {
        "_id": ObjectId('5c729df2e155ac16da86a1d0'),
        "catalog": ObjectId('5c710622e155ac0c39c8b66d'), # 所属分类节点ID, 即所属区块链
        "name": 'Simple Token',
        "symbol": 'TST',
        "decimals": 18,       # 有效位长度，即小数位数
        "precision": 6,       # 精度位数，在显示，交互中方便使用
        "icon": "https://cdn2.iconfinder.com/data/icons/5/100/cryptocurrency_blockchain_crypto-02-512.png",
        "type": "ERC20",     # 币种类型，如：main：主币（本系统将主币也作为一种特殊的合约），ERC20：ERC20币，ERC721：ERC721币
        "spec": {            # 主币spec为空
            "address": '0x7af963cF6D228E564e2A0aA0DdBF06210B38615D',  # 合约地址
            "data": [...],                                            # 合约相关数据，不同链存在不同的数据
        }
        "status": 10,          # -10:已删除，0: 休眠中，10: 活动中
        "created": 1551015331186,
        "updated": 1551015331186
    }

    本系统将主币也作为一种特殊的合约！！！
    """
    contract.create_index('spec.address')
    contract.create_index('symbol')
    contract.create_index('name')
    contract.create_index([('catalog', ASCENDING), ('created', DESCENDING)])


def _transaction_index():
    """
    {
        "_id": ObjectId('5c729df2e155ac16da86a1d0'),
        "chainName": "eth",
        "from": "0x58a81beab5f9948114d267ee834d82928accb030",
        "to": "0xd7fe1ac536e38831b160f587323fbd20c1b30daf",
        "hash": "0xc50aa6efbf1af2d92d585784f6a712e3d81c9447983c228a147a05fb9d3a176d",
        "blockHash": "0x21f0fdeb6ca0eccb554048c5793961b076067c53486cd8644e21973fd6e2cf79",
        "blockNumber": 425840,
        "transactionIndex": 3,
        "value": 0,
        "nonce": 163
    }

    本系统将主币也作为一种特殊的合约！！！
    """
    transaction.create_index([('from', ASCENDING), ('nonce', DESCENDING)], unique=True)
    transaction.create_index([('to', ASCENDING),
                              ('blockNumber', DESCENDING),
                              ('transactionIndex', DESCENDING)], unique=True)


def _tx_index():
    """
    {
        "_id": ObjectId('5c729df2e155ac16da86a1d0'),
        "status": 0,      # 0: 待广播，10：待确认，20：已确认
        "chainName": "eth",
        "from": "0x58a81beab5f9948114d267ee834d82928accb030",
        "to": "0xd7fe1ac536e38831b160f587323fbd20c1b30daf",
        "hash": "0xc50aa6efbf1af2d92d585784f6a712e3d81c9447983c228a147a05fb9d3a176d",
        "blockHash": "0x21f0fdeb6ca0eccb554048c5793961b076067c53486cd8644e21973fd6e2cf79",
        "blockNumber": 425840,
        "transactionIndex": 3,
        "value": 0,
        "nonce": 163
    }

    本系统将主币也作为一种特殊的合约！！！
    """
    tx.create_index([('from', ASCENDING), ('nonce', DESCENDING)])


def _init_data():
    pass



from webmother.passport import Passport
from tweb.license import License
from webmother.service import ctrl_passport

system_name = 'wallet'

max_lic_text = 'contract:11111;;'
profiles = {
    'contract': {
        'switch': [
            "create",
            "read",
            "remove",
            "activate",
            "deactivate"
        ]
    }
}
display = {
    'zh': {
        'contract': '合约管理',
        'contract.switch': '权限开关',
        'contract.switch.create': '创建',
        'contract.switch.read': '读取',
        'contract.switch.remove': '移除',
        'contract.switch.activate': '激活',
        'contract.switch.deactivate': '冻结',
    },
    'en': {
        'contract': 'Contract Manage',
        'contract.switch': 'Switches',
        'contract.switch.create': 'Create',
        'contract.switch.read': 'Read',
        'contract.switch.remove': 'Remove',
        'contract.switch.activate': 'Activate',
        'contract.switch.deactivate': 'Deactivate',
    }
}


def append_extra():
    # 添加本系统涉及到的权限项
    Passport.add_system_profile(system_name, profiles, display)

    # 更新超级管理员的账户相关权限
    extra_lic = License(profiles, display).parse(max_lic_text)
    ctrl_passport.simple_update('0', '0', extra_lic.json)



# coding=utf-8

from tweb import base_handler, myweb
from tornado import gen
import json
from ..service import ctrl_wallet as wallet


class VerifyHandler(base_handler.BaseHandler):

    # @myweb.authenticated
    @gen.coroutine
    def get(self, address, **kwargs):
        password = '123456'
        yield wallet.verify_password(address, password)
        return self.write({'result': 'pass'})


class MyWalletsHandler(base_handler.BaseHandler):

    @myweb.authenticated
    @gen.coroutine
    def get(self):
        user_id = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')

        ret = yield wallet.query_mine(user_id, access_token)
        return self.write({'list': ret})


class WalletHandler(base_handler.BaseHandler):
    """
    钱包账户操作
    """

    @myweb.authenticated
    @gen.coroutine
    def post(self, contract_id, wallet_id, **kwargs):
        user_id = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')

        data = json.loads(self.request.body.decode('utf-8'))
        ret = yield wallet.create(contract_id, data, user_id, access_token)
        return self.write(ret)

    @myweb.authenticated
    @gen.coroutine
    def delete(self, contract_id, wallet_id, **kwargs):
        user_id = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')

        ret = yield wallet.remove(contract_id, wallet_id, user_id, access_token)
        return self.write(ret)


class BalanceHandler(base_handler.BaseHandler):
    """
    钱包余额
    """

    @myweb.authenticated
    @gen.coroutine
    def get(self, contract_id, address, **kwargs):
        ret = yield wallet.balance(contract_id, address)
        return self.write(ret)


class TransferHandler(base_handler.BaseHandler):
    """
    交易转账（含合约交易）
    """

    @myweb.authenticated
    @gen.coroutine
    def post(self, contract_id, from_address, **kwargs):
        data = json.loads(self.request.body.decode('utf-8'))
        # ret = yield wallet.transfer(contract_id, from_address, data)
        ret = yield wallet.transfer_apply(contract_id, from_address, data)
        return self.write(ret)


class TransferEstimateHandler(base_handler.BaseHandler):
    """
    交易矿工费预估
    """

    @myweb.authenticated
    @gen.coroutine
    def post(self, contract_id, from_address, **kwargs):
        data = json.loads(self.request.body.decode('utf-8'))
        ret = yield wallet.transfer_miner_fei(contract_id, from_address, data)
        return self.write(ret)


class TxInfoByHashHandler(base_handler.BaseHandler):
    """
    获取交易详情（含合约交易，可查询链上所有交易）
    """

    @myweb.authenticated
    @gen.coroutine
    def get(self, cid, tx_hash, **kwargs):
        ret = yield wallet.tx_info(cid, tx_hash)
        return self.write(ret)


class TxInfoByIdHandler(base_handler.BaseHandler):
    """
    获取交易详情（只有在本系统提交的交易）
    """

    @myweb.authenticated
    @gen.coroutine
    def get(self, tx_id, **kwargs):
        ret = yield wallet.tx_details(tx_id)
        return self.write(ret)


class TxListHandler(base_handler.BaseHandler):
    """
    获取交易详情（含合约交易）
    """

    @myweb.authenticated
    @gen.coroutine
    def get(self, cid, address, **kwargs):
        ret = yield wallet.tx_list(cid, address)
        return self.write({'list': ret})


class ViewHandler(base_handler.BaseHandler):
    """
    合约查看（不用签名，不用矿工费，不会写入数据）
    """

    @myweb.authenticated
    @gen.coroutine
    def post(self, contract_id, **kwargs):
        data = json.loads(self.request.body.decode('utf-8'))
        ret = yield wallet.view(contract_id, data)
        return self.write(ret)


class TransactionHandler(base_handler.BaseHandler):
    """
    # 合约交易调用（需签名，会扣矿工费，会写入/改变数据）
    """

    @myweb.authenticated
    @gen.coroutine
    def post(self, contract_id, sender_address, **kwargs):
        data = json.loads(self.request.body.decode('utf-8'))
        ret = yield wallet.tx_apply(contract_id, sender_address, data)
        return self.write(ret)



from decimal import Decimal
import math

S_IDX = 0
P_IDX = 1
A_IDX = 2


def f2d(val):
    return Decimal(str(val))


class CoinNum(Decimal):
    uni_dec = 18

    def __init__(self, dec=18, prec=18):
        self.dec = dec
        self.prec = prec

    def set(self, dec=None, prec=None):
        if dec is not None:
            self.dec = dec
        if prec is not None:
            self.prec = prec
        return self

    def amount(self):
        return round(self)

    def precise(self):
        """
        精确小数
        :return: 字符串
        """
        return format(self / 10 ** self.dec, '.{}f'.format(self.dec))

    def simple(self, mode=None):
        """
        方便阅读的，待有效数字的小数
        :return: 字符串
        """
        if mode == 'up':
            delta = self.dec - self.prec
            up = f2d(math.ceil(self / 10 ** delta) * 10 ** delta)
            return format(up / 10 ** self.dec, '.{}f'.format(self.prec))
        elif mode == 'low':
            delta = self.dec - self.prec
            low = f2d(math.floor(self / 10 ** delta) * 10 ** delta)
            return format(low / 10 ** self.dec, '.{}f'.format(self.prec))
        else:
            return format(self / 10 ** self.dec, '.{}f'.format(self.prec))

    def uninum(self):
        """
        内部统一精度整数
        :return: int
        """
        return round(self * 10 ** (self.uni_dec - self.dec))

    def group(self, mode=None):
        return self.simple(mode), self.precise(), self.amount()

    @staticmethod
    def f2c(f_val, dec, prec):
        """
        将易读数字（浮点值或字符串）转为CoinNum对象
        :param f_val:
        :param dec: 小数点位数
        :param prec: 有效位数
        :return: CoinNum
        """
        return CoinNum(round(Decimal(str(f_val)) * 10 ** dec)).set(dec, prec)

    @staticmethod
    def u2c(uninum, dec, prec):
        """
        将统一精度数字转为CoinNum对象
        :param uninum:
        :param dec: 小数点位数
        :param prec: 有效位数
        :return: CoinNum
        """
        return CoinNum(round(uninum / 10 ** (CoinNum.uni_dec - dec))).set(dec, prec)

    @staticmethod
    def points(f_val):
        """
        获取小数后位数
        :param f_val: 浮点值或字符串
        :return: 小数位数
        """
        return abs(Decimal(str(f_val)).as_tuple().exponent)



from tweb import scheduler, time
from ...db import mongo as db
from ...service import ctrl_wallet, ctrl_contract
import asyncio
from ..external import ethereum as ex_ethereum


def start():
    interval = 1
    scheduler.add_interval(_tx_checking, interval)


async def _tx_checking():
    cursor = db.tx.find({
        'status': {'$in': [0, 10]}
    })

    pending_list = list()
    wallets = dict()

    # 按from地址分组
    for tx in cursor:
        if tx['status'] == 0:
            address = tx['from']
            if address not in wallets:
                wallets[address] = list()
            wallets[address].append(tx)
        elif tx['status'] == 10:
            pending_list.append(tx)

    # 对已广播的交易查询是否已在链上确认
    if len(pending_list) > 0:
        tasks = [_check(tx) for tx in pending_list]
        await asyncio.gather(*tuple(tasks))

    # 对未广播交易列表进行广播
    if len(wallets) > 0:
        tasks = [_cast(address, tx_list) for address, tx_list in wallets.items()]
        await asyncio.gather(*tuple(tasks))


async def _cast(sender_address, tx_list):
    nonce = ex_ethereum.get_tx_count(sender_address)

    last = db.tx.find_one({'from': sender_address, 'status': {'$in': [10, 20]}})
    if last is not None:
        last_nonce = last['nonce'] + 1
        nonce = max(nonce, last_nonce)

    for tx in tx_list:
        contract_id = tx['contract_id']
        data = tx['raw_data']

        tx_info = await ctrl_wallet.tx_cast(contract_id, sender_address, data, nonce=nonce)
        nonce += 1

        tx_info.pop('from')
        tx.update(tx_info)
        tx['status'] = 10
        tx['ts_cast'] = time.millisecond()
        db.tx.update_one({'_id': tx['_id']}, {'$set': tx})


async def _check(tx):
    block_number = ex_ethereum.block_number()
    contract = ctrl_contract.simple_read(tx['contract_id'])

    tx_info = await ctrl_wallet.tx_info(contract['catalog']['cid'], tx['hash'])

    block_no = tx_info.get('blockNumber')
    if block_no is not None and block_number - block_no >= 0:
        tx_info.pop('from')
        tx.update(tx_info)
        tx['status'] = 20
        tx['ts_mined'] = time.millisecond()
        db.tx.update_one({'_id': tx['_id']}, {'$set': tx})



# coding=utf-8

import config
from web3 import Web3
import json
from hexbytes import HexBytes
from tweb.error_exception import ErrException, ERROR
from tornado.options import options

web3 = None
chain = None


def init():
    global chain
    chain = options['chain']
    provider = Web3.HTTPProvider(config.Ethereum[chain]['http'], request_kwargs={'timeout': 60})

    global web3
    web3 = Web3(provider)


def wallet_balance(address):
    address = web3.toChecksumAddress(address)
    ret = web3.eth.getBalance(address)
    return str(ret)


def block_number():
    return web3.eth.blockNumber


def get_tx_count(address):
    address = web3.toChecksumAddress(address)
    return web3.eth.getTransactionCount(address)


def get_gas_price():
    return web3.eth.gasPrice


def send_raw_tx(raw_transaction):
    result = web3.eth.sendRawTransaction(raw_transaction)
    tx_hash = web3.toHex(result)
    return tx_hash


def create_tx(from_address, to_address, value, nonce=None):
    from_address = web3.toChecksumAddress(from_address)
    to_address = web3.toChecksumAddress(to_address)

    if nonce is None:
        nonce = get_tx_count(from_address)
    gas_price = get_gas_price()

    transaction = {
        'from': from_address,
        'to': to_address,
        'value': int(value),
        'gasPrice': gas_price,
        # 'data': "",
        'nonce': nonce,
        # 'chainId': 5
    }

    gas = web3.eth.estimateGas(transaction)
    transaction['gas'] = gas

    return transaction


def get_tx(tx_hash):
    tx = web3.eth.getTransaction(tx_hash)
    if tx is None:
        raise ErrException(ERROR.E40400)
    json_str = json.dumps(dict(tx), cls=HexJsonEncoder)
    ret = json.loads(json_str)
    return ret


# 合约调用相关


def contract_call(contract_json, method, args=None):
    contract_address = web3.toChecksumAddress(contract_json['spec']['address'])
    abi = contract_json['spec']['data']
    contract = web3.eth.contract(contract_address, abi=abi)

    if args is None:
        args = []

    temp = list()
    for i, val in enumerate(args):
        if web3.isAddress(val):
            temp.append(web3.toChecksumAddress(val))
        else:
            temp.append(val)

    try:
        ret = contract.functions[method](*tuple(temp)).call()
        if isinstance(ret, int):
            ret = str(ret)
    except Exception as e:
        raise ErrException(ERROR.E50000, extra='%s' % e)

    return ret


def contract_create_tx(contract_json, method, args, tx_param):
    contract_address = web3.toChecksumAddress(contract_json['spec']['address'])
    abi = contract_json['spec']['data']

    contract = web3.eth.contract(contract_address, abi=abi)

    if method not in contract.functions:
        raise ErrException(ERROR.E40000, extra='the contract not support the method %s' % method)

    temp = list()
    for i, val in enumerate(args):
        if web3.isAddress(val):
            temp.append(web3.toChecksumAddress(val))
        else:
            temp.append(val)

    try:
        transaction = contract.functions[method](*tuple(temp)).buildTransaction(tx_param)
    except Exception as e:
        raise ErrException(ERROR.E50000, extra='%s' % e)

    gas = web3.eth.estimateGas(transaction)
    transaction['gas'] = gas

    return transaction


# 其他辅助工具
class HexJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, HexBytes):
            return obj.hex()
        return super().default(obj)



# coding=utf-8

from tweb import base_handler, myweb
from tornado import gen
from webmother.service.async_wrap import ctrl_entrance
from ..extra_passport import system_name


class EmployPassportsHandler(base_handler.BaseHandler):
    """
    获取我们(组织)的授权列表
    """

    @myweb.authenticated
    @gen.coroutine
    def get(self,  **kwargs):
        employment = self.request.headers.get('x-signed-employment')
        access_token = self.request.headers.get('x-access-token')

        array, display = yield ctrl_entrance.employ_passports(system_name, employment, access_token)

        return self.write({'list': array, 'passport_display': display})
